// sifilm.c
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "common.h"
#include "camera.h"
#include "math3d.h"
#include "buffer.h"
#include "oglview.h"
#include "binfile.h"
#include "sifilm.h"
#include "sires.h"
#include "log.h"

#if 0
processslice
	si=t_nextvar
	di=t_prevvar
	if(tapedir?)
			t_beforeflags
			t_nextbeforeflags
			loop
				if((dx>1) nc)
					[si-2]=t_beforevar-t_prevar-w[si]
	else
		t_nextbeforeflags
		t_beforeflags

	nextslice
#endif

SIFILM *film;

static int verbose=TRUE;

static unsigned char get_byte(SIFILM *f)
{
	return(f->tape[f->pos++]);
}
static unsigned short get_word(SIFILM *f)
{
unsigned short w;
	w=(unsigned short)(f->tape[f->pos])|((unsigned short)(f->tape[f->pos+1])<<8);
	f->pos+=2;
	return(w);
}

	//larges are diffrent from ints. int 1,2,3,4 large 3,4,1,2
static unsigned int get_large(SIFILM *f)
{
unsigned int i,j;
	j=(uint)get_word(f);
	i=(uint)get_word(f);

	i=(i<<16)|j;
	return(i);
}

int sifilm_load(char *name)
{
int i;
SIFILM *f;
FILE *fi;
char sig[8];

	Binfile file(name);
	if(!file.test() )
		return(NULL);
	
	file.read(sig,4);
	if( !strstr(sig,"FILM") )		 //check file signature
		return(NULL);
	if(verbose)
		LogPrintf("Decoding FLM:%s\n",name);

	f=(SIFILM *)malloc( sizeof(SIFILM) );
	if(f==NULL)
		return(NULL);	//not enuf mem.


	f->len=file.size()-4; //-4 bypasses header
	f->tape=(uchar *)malloc(f->len);
	if(f->tape==NULL)
		{
		free(f->tape);
		return(NULL);	//not enuf mem.
		}
	file.read((char *)f->tape,f->len);

	f->pos=0;//used as index into buf (4=bypass 'FILM' in file)
	f->max_obj=0;
	f->view_obj=0;
	for(i=0;i<SIFILM_MAX_OBJ;i++)
		{
		f->objs[i].type=0;
		f->objs[i].flags=0;
		}
	film=f;
	return(TRUE);
}


void process_slice(SIFILM *f)
{
unsigned char c;
SIFILM_OBJ *obj;
unsigned short w;
int i;

	c=get_byte(f);	//inctime.
	w=get_word(f);
//	LogPrintf("\nSLICE start:%08x LEN:%02x:%04x\n",f->pos+9,c,w);

	while(f->pos<f->len)	//loop until end of tape
		{
		c=get_byte(f);
		if(c==0xff)
			return;		//end of slice
		switch( c>>6 )
			{
			case 0:
				obj=&f->objs[c&0x3f];
				obj->type=get_byte(f);
//            LogPrintf("%02x:CREATE: #%02x\n",f->pos-1,c&0x3f);
				if(obj->type&SIF_EXISTS)
					{
					if(obj->type&SIF_DRAW)
						obj->shape=get_word(f);
					if(!(obj->type&SIF_MOVES))
						{
						obj->x=get_large(f);
						obj->y=get_large(f);
						obj->z=get_large(f);
						}
					if(!(obj->type&SIF_ROTATES))
						{
						obj->rz=get_word(f);
						obj->rx=get_word(f);
						obj->ry=get_word(f);
						}
					for(i=0;i<6;i++) //might not be needed
						obj->data1[i]=0;
					}
				else
					{
//	            LogPrintf("   CREATE NON_EXIST:\n");
					//exit(0);
					}
				break;
			case 1:
				obj=&f->objs[c&0x3f];
//            LogPrintf("%02x:INTERP: #%02x\n",f->pos,c);
				if(obj->type&SIF_EXISTS)
					{
					if(obj->type&SIF_MOVES)
						{
						obj->data1[0]=get_large(f);
						obj->data1[1]=get_large(f);
						obj->data1[2]=get_large(f);
						}
					if(obj->type&SIF_ROTATES)
						{
						obj->data1[3]=get_word(f);
						obj->data1[4]=get_word(f);
						obj->data1[5]=get_word(f);
						}
					if(obj->type&SIF_ANIMATES)
						obj->animate=get_word(f);
					}
				else
					{
	            LogPrintf("ERROR:   INTERP NON_EXIST:\n");
					exit(0);
					}
				break;
			case 2:
//				LogPrintf("MISC\n");
				if(c<0x90) //1001000
					{
					//based on tapedir sets either prevar or nextvar
					f->nextvar[c&0x0f]=get_word(f);
//					LogPrintf("   %02x:VARIBLE:#%02x -->%04x\n",f->pos,c&0xf,w);
					}
				else if(c<0x98) //1001 1000
					{
					f->view_obj=(c&0xf);
//					LogPrintf("   %02x:VIEW:-->%02x\n",f->pos,f->view_obj);
					}
				else if(c==0x98) //1001 1000
					{
					f->credits[0]=0;
//					LogPrintf("   %02x:CLEAR CREDIT:\n",f->pos);
					}
				else if(c==0x99) //1001 1001
					{
//					LogPrintf("   %02x:CREDIT:%s\n",f->pos,&f->tape[f->pos]);
					i=0;
					while((f->credits[i++]=get_byte(f))!=0);
					while((f->credits[i++]=get_byte(f))!=0);
					}
				else if(c==0x9b) //1001 1011
					{
					f->max_obj=get_byte(f);
//					LogPrintf("   %02x:MAXOBJ:-->%02x\n",f->pos,f->max_obj);
					}
				else if(c==0x9c) //1001 1100
					{
					get_byte(f);
					get_word(f);
//					LogPrintf("   %02x:SOUND:%02x\n",f->pos);
//					exit(0);
					}
				else if(c>=0xa0) //1010 0000
					{
					w=get_word(f);
//					LogPrintf("   %02x:BEFORE:%02x\n",f->pos,w);
					}
				else
					{
					LogPrintf("ERROR:  UNK %08x\n",f->pos);
					exit(0);
					}
				break;
			case 3:
				LogPrintf("??UNK %08x\n",f->pos);
				exit(0);
				break;
			}
		}
//	exit(0);
}
void sifilm_dump_state(SIFILM *f)
{
int i;
SIFILM_OBJ *o;

	LogPrintf("---------------------------------------------------------\n");
	LogPrintf("FILM: POS:%08x LEN:%08x MAXOBJ:%02x VIEWOBJ:%02x\n",
		f->pos,f->len,f->max_obj,f->view_obj);
	for(i=0;i<f->max_obj;i++)
		{
		o=&f->objs[i];
		if(o->type&SIF_EXISTS)
			{
			LogPrintf(" OBJ#%02x FL:%02x TYPE:",i,o->flags);
				//print flags
			if(o->type&SIF_EXISTS)LogPrintf("E");else LogPrintf(" ");
			if(o->type&SIF_DRAW)LogPrintf("D");else LogPrintf(" ");
			if(o->type&SIF_SHADOW)LogPrintf("S");else LogPrintf(" ");
			if(o->type&SIF_MOVES)LogPrintf("M");else LogPrintf(" ");
			if(o->type&SIF_ROTATES)LogPrintf("R");else LogPrintf(" ");
			if(o->type&SIF_ANIMATES)LogPrintf("A ");else LogPrintf("  ");

			if(o->type&SIF_MOVES)
				LogPrintf(" VPOS:%08x,%08x,%08x ",o->data1[0],o->data1[1],o->data1[2]);
			else
				LogPrintf(" sPOS:%08x,%08x,%08x ",o->x,o->y,o->z);
			if(o->type&SIF_ROTATES)
				LogPrintf(" VROT:%04x,%04x,%04x ",o->data1[3],o->data1[4],o->data1[5]);
		  	else
				LogPrintf(" ROT:%04x,%04x,%04x ",o->rx,o->ry,o->rz);
//	LogPrintf(" ROT:%04x,%04x,%04x ",o->rx,o->ry,o->rz);
			if(o->type&SIF_DRAW)
				LogPrintf("SHAPE:%04x",o->shape);
			LogPrintf("\n");
			}
		}
	LogPrintf("---------------------------------------------------------\n");
}



void sifilm_play()
{
	if(film->pos>=film->len)
		film->pos=0;

	process_slice(film);

		// this doesnt work yet.
   //sifilm_set_camera(film);
}
void sifilm_draw(OGLview *v)
{
int i;
SIFILM_OBJ *o;
float x,y,z;
float rx,ry,rz;
SIFILM* f=film;
int scale=0x10000;
float rs=0x10000/360;

	for(i=0;i<f->max_obj;i++)
		{
		o=&f->objs[i];
		if(o->type&SIF_EXISTS)
			{
			if(o->type&SIF_DRAW)
				{
				glPushMatrix();
				if(o->type&SIF_MOVES)
					{
					x=(float)o->data1[0]/scale;
					y=(float)o->data1[1]/scale;
					z=(float)o->data1[2]/scale;
					}
				else
					{
					x=(float)o->x/scale;
					y=(float)o->y/scale;
					z=(float)o->z/scale;;
					}
				if(o->type&SIF_ROTATES)
					{
					rx=((float)o->data1[3])/rs;
					ry=((float)o->data1[4])/rs;
					rz=((float)o->data1[5])/rs;
					}
				else
					{
					rx=((float)o->rx)/rs;
					ry=((float)o->ry)/rs;
					rz=((float)o->rz)/rs;
					}
				glTranslatef(-x,-y,z);
				glRotatef(rx, 1.0f, 0.0f, 0.0f); /* rotate about the x unit vector */

				glRotatef(rz, 0.0f, 1.0f, 0.0f); /* rotate about the y unit vector */

				glRotatef(ry, 0.0f, 0.0f, 1.0f); /* rotate about the z unit vector */


				sires_draw(o->shape);
				glPopMatrix();
				}

//			sifilm_dump_state(f);
			}
		}
}
void sifilm_test()
{
SIFILM *f;

//	f=sifilm_load("szpromo.flm");
//	f=sifilm_load("bob1.flm");
//	f=sifilm_load("pittjet1.flm");

	film=f;

	if(f!=NULL)
		{
		sifilm_dump(f);
		}
	exit(0);
}


