// simap.c

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "common.h"
#include "oglview.h"
#include "mesh.h"
#include "buffer.h"
#include "binfile.h"
#include "log.h"
#include "simap.h"
#include "sires.h"
#include "sisland.h"

typedef struct{
	int x,y,z;
	}SIPOINT;

typedef struct{
	int type;
	int x,y,z;
	unsigned int next;	 //index
	}POS_TYPE;

typedef struct{
	int type;
	int x,z;
	unsigned int next;  //index
	}POSXZ_TYPE;

typedef struct{
	int type;
	int obj;
	int ext;
	int splitdist;
	unsigned int next;  //index
	}NODE1_TYPE;

typedef struct{
	int type;
	unsigned int ptr;  //index
	unsigned int next;  //index
	}NODE2_TYPE;


typedef struct{
	int type;
	unsigned int ptr;  //index
	unsigned int next;  //index
	int sortvalue;  //??
	}XPLANE_TYPE;


typedef struct{
	int type;
	unsigned int ptr;  //index
	unsigned int next;  //index
	int sortvalue;  //??
	}ZPLANE_TYPE;

typedef struct{
	int type;
	unsigned int ptr;	 //index
	unsigned int next;	 //index
	int slope;	
	int coord;	
	}HSLOPE_TYPE;


typedef struct{
	int type;
	unsigned int ptr;	 //index
	unsigned int next;	 //index
	int slope;	
	int coord;	
	}VSLOPE_TYPE;


typedef struct{
	int type;
	int	obj;
	}LEAF_TYPE;

typedef struct{
	int type;
	unsigned int next;	 //index
	//poly here.    lgl_poly	dw	?
	}GLEAF_TYPE;

typedef struct{
	int type;
	int splitdist;
	int ext;
	int x,y,z;
	unsigned int ptr;  //index
	unsigned int next; //index
	//poly here.    lgr_poly	dw	?
	}GROUND_TYPE;


Buffer nodes_buf;
static int *nodes;
static void simap_render_node(int ni,SIPOINT *cp,SIPOINT *ep);
static Mesh *island;

int simap_load(char *name)
{
FILE *f;
int size;
int *n;


	if(name==NULL)
		{	//load the default scenery file
		if(!nodes_buf.load_file("sidata\\scenery.scn"))
			return(FALSE);
		}
	else
		{ //load the specified scenery file
		if(!nodes_buf.load_file(name))
			return(FALSE);
		}

	nodes=(int *)&nodes_buf.data[0];

		//load the island and ocean
	island=island_load();

	return(TRUE);
}


int aprox_dist(int x1, int y1, int x2, int y2)
{
int dist;

//	dist=(int)vect_dist((float)x1,(float)y1,(float)x2,(float)-y2);
//	return(dist);
/*
 * gives approximate distance from (x1,y1) to (x2,y2)
 * with only overestimations, and then never by more
 * than (9/8) + one bit uncertainty.
 */
	if ((x2 -= x1) < 0) x2 = -x2;
	if ((y2 -= y1) < 0) y2 = -y2;
	return (x2 + y2 - (((x2>y2) ? y2 : x2) >> 1) );
}

	//Draw the object by index number at Point P.
static void draw_object(int oi,SIPOINT *p)
{
int i;
	if(oi==0)
		return;

	if(oi==0x328)
		return;

	glPushMatrix();
	glTranslatef((float)-p->x,(float)-p->y,(float)p->z);
	sires_draw(oi);
	glPopMatrix();

//	LogPrintf("Object:%x X:%04x Y:%04x Z:%04x\n",oi,p->x,p->y,p->z);
}


void simap_render()
{
SIPOINT p;
SIPOINT e;

	if(island)
		current_view->draw(island);

	p.x=(int)0;
	p.y=(int)0;
	p.z=(int)0;

	Vector3 v=current_view->camera.v_origin();

	e.x=(int)v.x;
	e.y=(int)v.y;
	e.z=(int)v.z;
	simap_render_node(1,&p,&e);
}


static void simap_render_node(int ni,SIPOINT *cp,SIPOINT *ep)
{
SIPOINT np;
int i;
int split_dist;
int oi;
POS_TYPE *pos_type;
POS_TYPE *posxz_type;

	while(1)
		{
		if(ni==0)
			return;
		switch(nodes[ni])
			{
			case SP_POS:
				pos_type=(POS_TYPE *)&nodes[ni];
   			cp->x+=(long)((short)(((POS_TYPE *)&nodes[ni])->x));
   			cp->y+=(long)((short)(((POS_TYPE *)&nodes[ni])->y));
   			cp->z+=(long)((short)(((POS_TYPE *)&nodes[ni])->z));
   			ni=((POS_TYPE *)&nodes[ni])->next/2;
				break;
			case SP_POSXZ:
   			cp->x+=(long)((short)(((POSXZ_TYPE *)&nodes[ni])->x));
   			cp->z+=(long)((short)(((POSXZ_TYPE *)&nodes[ni])->z));
   			ni=((POSXZ_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_1NODE:
				//draw obj
				//minimum lod?
				split_dist=(( (NODE1_TYPE *)&nodes[ni])->splitdist );
            oi=((NODE1_TYPE *)&nodes[ni])->obj;

				if(aprox_dist(cp->x,-cp->z,ep->x,ep->z)>=split_dist*48)
					{
					if(oi)
						{
						return;
						}
					}
   			ni=((NODE1_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_2NODE:
				np.x=cp->x;np.y=cp->y;np.z=cp->z;
				simap_render_node( ((NODE2_TYPE *)&nodes[ni])->ptr/2,&np,ep );
   			ni=((NODE2_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_XPLANE:
		   case SP_ZPLANE:
				np.x=cp->x;np.y=cp->y;np.z=cp->z;
				simap_render_node( ((XPLANE_TYPE *)&nodes[ni])->ptr/2,&np,ep );
   			ni=((XPLANE_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_HSLOPE:
		   case SP_VSLOPE:
				np.x=cp->x;np.y=cp->y;np.z=cp->z;
				simap_render_node( ((HSLOPE_TYPE *)&nodes[ni])->ptr/2,&np,ep );
   			ni=((VSLOPE_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_LEAF:
				draw_object( (( (LEAF_TYPE *)&nodes[ni])->obj ),cp );
				ni=0;
				break;
		   case SP_GROUND:
				np.x=cp->x;np.y=cp->y;np.z=cp->z;
				simap_render_node( ((GROUND_TYPE *)&nodes[ni])->ptr/2,&np,ep );
   			ni=((GROUND_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_GLEAF:
   			ni=((GLEAF_TYPE *)&nodes[ni])->next/2;
				break;
		   case SP_LIST:
				ni++;
				while(nodes[ni])
					{
					np.x=cp->x;np.y=cp->y;np.z=cp->z;
					simap_render_node( nodes[ni]/2,&np,ep );
	   			ni++;
					}
   			ni++;
				break;
			default:
				LogPrintf("Script Error");
				exit(-1);
			}
		}
}










