#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "common.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include "mesh.h"
#include "sires.h"
#include "sipal.h"
#include "siobj.h"
#include "sod.h"
#include "buffer.h"
#include "log.h"

SIVERTEX verts[2000];
SIVERTEX cur_vert;
static float scale=1.0;
static int indent=0;
static int yoff;
static int verbose=FALSE;

//todo
// add option to auto remove detail polys
// support lines block circle and sphere in that order.


	//C versions of the model opcodes
static void set_vert(SIVERTEX *v,int x,int y,int z)
{
	v->x=x;v->y=y;v->z=z;
	v->flags=TRUE;
}

static void get_vert(SIVERTEX *d,SIVERTEX *s)
{
	d->x=s->x;
	d->y=s->y;
	d->z=s->z;
	if(!s->flags)
		LogPrintf("Getting for Un-inited vert");

	d->flags=TRUE;

}
static void mid_vert(SIVERTEX *d,SIVERTEX *s)
{
SIVERTEX t;

	if(!s->flags || !d->flags)
		LogPrintf("Miding Un-inited vert");

	t.x=(s->x-d->x)/2;
	t.y=(s->y-d->y)/2;
	t.z=(s->z-d->z)/2;
	
	d->x+=t.x;
	d->y+=t.y;
	d->z+=t.z;

}

static void add_vert(SIVERTEX *d,SIVERTEX *s)
{
SIVERTEX t;

	if(!s->flags || !d->flags)
		LogPrintf("Miding Un-inited vert");

	t.x=(s->x+d->x);
	t.y=(s->y+d->y);
	t.z=(s->z+d->z);
	
	d->x=t.x;
	d->y=t.y;
	d->z=t.z;

}


static void sub_vert(SIVERTEX *d,SIVERTEX *s)
{
SIVERTEX t;

	if(!s->flags || !d->flags)
		LogPrintf("Miding Un-inited vert");

	t.x=(d->x-s->x);
	t.y=(d->y-s->y);
	t.z=(d->z-s->z);
	
	d->x=t.x;
	d->y=t.y;
	d->z=t.z;

}
static void neg_vert(SIVERTEX *v)
{
	if(!v->flags)
		LogPrintf("neging Un-inited vert");
	v->x=-v->x;
	v->y=-v->y;
	v->z=-v->z;
}
static void shr_vert(SIVERTEX *v,int s)
{
	if(!v->flags)
		LogPrintf("neging Un-inited vert");
	v->x=v->x>>s;
	v->y=v->y>>s;
	v->z=v->z>>s;
}
static void shl_vert(SIVERTEX *v,int s)
{
	if(!v->flags)
		LogPrintf("neging Un-inited vert");
	v->x=v->x<<s;
	v->y=v->y<<s;
	v->z=v->z<<s;
}

static void dump_vert(SIVERTEX *v)
{
	if(!v->flags)
		LogPrintf("Dumping Un-inited vert");
	LogPrintf("--->%d,%d,%d",v->x,v->y,v->z);
}

void siobj_test()
{
//	siobj_load("temp");
	exit(-1);
}




int siobj_parse_ipg(ushort *buf,int bi)
{
int vert_index=0;
ushort w,w1,w2,w3,w4;

   w=buf[bi++];
	switch(w)
		{
		case 0:		//make a cube starting at vertex in w4
     		w1=buf[bi++];
  	      w2=buf[bi++];
       	w3=buf[bi++];
       	w4=buf[bi++]/0xc;
  			vert_index=w4;
  			//LogPrintf("IPG CUBE <%d,%d,%d>,%d\n",w,w1,w2,w3,w4);
  			set_vert(&verts[vert_index+7],0+w1,0+w2,0-w3);
  			set_vert(&verts[vert_index+6],0+w1,0+w2,0+w3);
  			set_vert(&verts[vert_index+5],0-w1,0+w2,0+w3);
  			set_vert(&verts[vert_index+4],0-w1,0+w2,0-w3);
  			set_vert(&verts[vert_index+3],0+w1,0-w2,0-w3);
  			set_vert(&verts[vert_index+2],0+w1,0-w2,0+w3);
  			set_vert(&verts[vert_index+1],0-w1,0-w2,0+w3);
  			set_vert(&verts[vert_index+0],0-w1,0-w2,0-w3);
  		  	break;
  		case 2:  //make a square starting at vertex in w4
      	w1=buf[bi++];
       	w2=buf[bi++];
       	w4=buf[bi++]/0xc;
  			vert_index=w4;
  			//LogPrintf("IPG GROUNDPTS <%d,%d>,%d\n",w,w1,w2,w4);
  			set_vert(&verts[vert_index+3],0+w1,0,0-w2);
  			set_vert(&verts[vert_index+2],0+w1,0,0+w2);
  			set_vert(&verts[vert_index+1],0-w1,0,0+w2);
  			set_vert(&verts[vert_index+0],0-w1,0,0-w2);
  		  	break;
  		case 8:  //make a square starting at vertex in w4
      	w1=buf[bi++];
       	w2=buf[bi++];
       	w4=buf[bi++]/0xc;
  			vert_index=w4;
  			//LogPrintf("IPG FLAT <%d,%d>,%d\n",w,w1,w2,w4);
  			set_vert(&verts[vert_index+3],0+w1,0-w2,0);
  			set_vert(&verts[vert_index+2],0+w1,0+w2,0);
  			set_vert(&verts[vert_index+1],0-w1,0+w2,0);
  			set_vert(&verts[vert_index+0],0-w1,0-w2,0);
  		  	break;
  		default:
      	w1=buf[bi++];
  			LogPrintf("IPG type wrong:%d @ %x -->%x\n",w,bi,w1);
  		  	break;
  		}
	return(bi);
}


int cur_color=0;

int siobj_parse_obj(short *buf,int bi)
{
short w,w1,w2,w3,w4,w5;
int i;
int num;
int size;

	while(1)	//loop 'till end of file
		{
		w=buf[bi++];
		switch(w)
			{
			case OP_IPG:	 //initilize vertex cache
				bi=siobj_parse_ipg((ushort*)buf,bi);
				if(bi==0)
					return(FALSE);
				break;
			case OP_VECTORS:
				w=buf[bi++];			//number of verts
				w1=buf[bi++]/0xc;		//start register.
				//LogPrintf("VECTORS #%d Start:%d\n",w,w1);
				for(i=0;i<w;i++)
					{
					w2=buf[bi++];
					w3=buf[bi++];
					w4=buf[bi++];

					//HMM not sure why the -w3 clipper mast bug fix
               set_vert(&verts[w1],w2,-w3,w4);
					w1++;
					}
				break;
			case OP_GETV:
				w=buf[bi++]/0xc;
				get_vert(&cur_vert,&verts[w]);
				break;
			case OP_INTV:
				w=buf[bi++]/0xc;
				mid_vert(&cur_vert,&verts[w]);
				break;
			case OP_PUTV:
				w=buf[bi++]/0xc;
				get_vert(&verts[w],&cur_vert);
				break;
			case OP_ADDV:
				w=buf[bi++]/0xc;
				add_vert(&cur_vert,&verts[w]);
				break;
			case OP_SUBV:
				w=buf[bi++]/0xc;
				sub_vert(&cur_vert,&verts[w]);
				break;
			case OP_SHRV:
				w=buf[bi++];
				shr_vert(&cur_vert,w);
				break;
			case OP_SHLV:
				w=buf[bi++];
				shl_vert(&cur_vert,w);
				break;
			case OP_NEGV:
				neg_vert(&cur_vert);
				break;
			case OP_CLRV:
				set_vert(&cur_vert,0,0,0);
				break;
			case OP_PERS:
				w=buf[bi++];
				w1=buf[bi++];
//				LogPrintf("PERSPEC %d,%d\n",w1,w);
				break;
			case OP_REJ:
				w=buf[bi++];
//				LogPrintf("REJ3D %d\n",w);
				w=buf[bi++];
//				LogPrintf("Num Faces:%d\n",w);
				for(i=0;i<w;i++)
					{
           		w1=buf[bi++];
           		w2=buf[bi++];
           		w3=buf[bi++];
//					LogPrintf("face <%d,%d,%d>\n",w1/0xc,w2/0xc,w3/0xc);
					}
				break;
			case OP_VEC:
				LogPrintf("ERROR OP_VEC");
				exit(-1);
        		w1=buf[bi++];
           	w2=buf[bi++];
           	w3=buf[bi++];
				//LogPrintf("vec <%d,%d,%d>\n",(int)w1,(int)w2,(int)w3);
				break;
			case OP_SPHERE:  //primitive not supported
        		w1=buf[bi++];
        		w2=buf[bi++];
				//LogPrintf("SPHERE %d,%d\n",(int)w1,(int)w2);
				break;
			case OP_BLOCK:    //primitive not supported
        		w1=buf[bi++];
        		w2=buf[bi++];
				//LogPrintf("BLOCK %d,%d\n",(int)w1,(int)w2);
				break;
			case OP_PERFDISK: //primitive not supported
        		w1=buf[bi++];
        		w2=buf[bi++];
        		w3=buf[bi++];
				//LogPrintf("PERFDISK %d,%d,%d\n",(int)w1,(int)w2,(int)w3);
				break;
			case OP_IFVIS:
				w=buf[bi++];
//				LogPrintf("IFVIS %d\n",w);
				w=buf[bi++]; //section size
//				LogPrintf("Size %x:%d\n",w);
				break;
			case OP_COLOR:
				w=buf[bi++];
				cur_color=w;
//				LogPrintf("FILLCOLOR %d\n",w);
				break;
			case OP_LINECOL:
				w=buf[bi++];
//				LogPrintf("LINECOL %d\n",w);
				break;
			case OP_MOVETO:
				w=buf[bi++];
//				LogPrintf("MOVETO %d\n",w/0xc);
				break;
			case OP_ZLINETO:
				w=buf[bi++];
//				LogPrintf("ZLINETO %d\n",w/0xc);
				break;
			case OP_DOT:
				w=buf[bi++];
//				LogPrintf("DOT %d\n",w/0xc);
				break;
			case OP_POLYZ:
//				LogPrintf("POLYZ ");
//				LogPrintf("\n");
				w=buf[bi++];
				//LogPrintf("%d,",w);
				//LogPrintf("<");

				glColor3f((float)sipal[(cur_color*3)+0]*(1.0/255.0),
						(float)sipal[(cur_color*3)+1]*(1.0/255.0),
						(float)sipal[(cur_color*3)+2]*(1.0/255.0) );
				glBegin(GL_POLYGON);		//use this for filled polys
//				glBegin(GL_LINE_LOOP);	//use this for Wire frame
				for(i=0;i<w;i++)
					{
           		w1=buf[bi++]/0xc;
               glVertex3f( (float)verts[w1].x*scale,
									(float)(verts[w1].y*scale)+(yoff*scale),
									(float)verts[w1].z*scale);
					}
				glEnd();
				break;
			case OP_GVPOLY16:
			case OP_GVPOLY:
//				LogPrintf("GVPOLY\n");
				num=buf[bi++];
				//LogPrintf("%d,%d",num,w);
				//LogPrintf("<");

				glColor3f((float)sipal[(cur_color*3)+0]*(1.0/255.0),
						(float)sipal[(cur_color*3)+1]*(1.0/255.0),
						(float)sipal[(cur_color*3)+2]*(1.0/255.0) );
				glBegin(GL_POLYGON); //use this for filled polys
//				glBegin(GL_LINE_LOOP); //use this for Wire frame
				for(i=0;i<num;i++)
					{
           		w1=buf[bi++]/0xc;
               glVertex3f( (float)verts[w1].x*scale,
									(float)(verts[w1].y*scale)+(yoff*scale),
//									(float)(verts[w1].y+yoff)*scale,
									(float)verts[w1].z*scale);
					}
				glEnd();
				w=buf[bi++]/0xc;
					//don't know/remember what this is.
				//LogPrintf(">,%d,<",w);
				for(i=0;i<num;i++)
					{
           		w1=buf[bi++];
					//LogPrintf("%d,",w1/2);
					}
				//LogPrintf(">\n");
				break;
			case OP_SHADES:
				w=buf[bi++];
//				LogPrintf("SHADES %d\n",w);
				w=buf[bi++];
				for(i=0;i<w;i++)
					{
	        		w1=buf[bi++];
   	        	w2=buf[bi++];
   	        	w3=buf[bi++];
//					LogPrintf("vec <%d,%d,%d>\n",(int)w1,(int)w2,(int)w3);
					}
				break;
			case 0x18c:
				w1=buf[bi++];
				w2=buf[bi++];
				w3=buf[bi++];
//				LogPrintf("SORT %x %x %x\n",(bi+(w1/2))-3,bi+(w2/2)-2,bi+(w3/2)-2);
					//gosub to next instruction
				if(siobj_parse_obj(buf,bi)==FALSE)
					return(FALSE);
					//gosub to second arg.
				if(siobj_parse_obj(buf,bi+(w2/2)-2)==FALSE)
					return(FALSE);
					//set pc to first arg.
            bi=(bi+(w1/2))-3;
				break;
			case 0x7e:	//push center
				LogPrintf("PUSH CENTER\n");
				break;
			case 0x87:	//pop center
				LogPrintf("POP CENTER\n");
				break;
			case 0x90:	//move center
				w1=buf[bi++];
				LogPrintf("MOVE CENTER: %d\n",w1);
				break;
			case OP_RETURN:
//				LogPrintf("RETURN:%x\n",bi);
				return(bi);
				break;

			case 0x1a7:	//gosub
				w1=buf[bi++];
//				LogPrintf("GOSUB:%x\n",w1);
				if(siobj_parse_obj(buf,bi+(w1/2)-1)==FALSE)
					return(FALSE);
				break;
			case 0x19e:	//goto
				w1=buf[bi++];
				LogPrintf("GOTO:%x\n",w1);
				bi=(bi+(w1/2)-1);
//					return(FALSE);
				break;
			case 0x22e:	//TAP whatever the fuck that is...
				LogPrintf("ACK: OP_TAP! Aborting.\n");
				return(FALSE);
				break;
#if 1
			case 0x2fd:	//bvgt
				w1=buf[bi++];
				w1=buf[bi++];
				w1=buf[bi++];
				break;
			case 0x2e2:	//rotx
            LogPrintf("ROTX:%x\n",buf[bi++]);
				break;
			case 0x2eb:	//rotz
            LogPrintf("ROTX:%x\n",buf[bi++]);
				break;
			case 0x144:	//shade1
				w1=buf[bi++];
				w2=buf[bi++];
            LogPrintf("SHADE1:%x %x\n",w1,w2);
				break;
			case 0x2b5:	//shade16
				w1=buf[bi++];
				w2=buf[bi++];
            LogPrintf("SHADE16:%x %x\n",w1,w2);
				break;
			case 0x1e6:	//pushrelmat
            LogPrintf("PUSHRELMAT:\n");
				break;
			case 0x1ef:	//poprelmat
            LogPrintf("POPRELMAT:\n");
				break;
#endif
			default:
				w1=buf[bi++];
				LogPrintf("ERROR Unknown op:%x:%d next %x:%d\n",w,w,w1,w1);
				LogPrintf("OPDESC:%s\n",sod_op_desc[w/0x9]);
				return(FALSE);
				break;

			}

		}
}

int siobj_parse(Buffer &buf)
{
int flags;
short w;
int i;

//	LogPrintf("Starting parse at:%08x\n",buf.pos);

	w=buf.next_word();
	if(w!=OP_ID)
		{
		LogPrintf("ERROR:Id not found:%x\n",w);
		return(FALSE);
		}

	w=buf.next_word(); 	//dunno what this word is
	if(verbose)
		LogPrintf("SHAPESIZE %x:%d\n",w,w);

	w=buf.next_word();  //Y off ground.
	yoff=w;
	if(verbose)
		LogPrintf("Yoff:%d\n",yoff);

	w=buf.next_word(); //scale and flags in this word.
	flags=w>>8;	 //set flags
	w=w&0xff;	 //mask off flags

	if(verbose)
		if(flags)	  //only flag is 1 meaning FLAT
			LogPrintf("Flags: %x\n",flags);

	if(verbose)
		LogPrintf("Scale: %x:%d\n",w,w);
	
	if(w>17)
		LogPrintf("ERROR: Scale out of range:%d\n",w);

	scale=scale_table[w]/10;
//scale=.001;

	w=buf.next_word(); //collision. whatever that is?
	if(verbose)
		if(w)
			LogPrintf("Collision %x:%d\n",w,w);

		//clear out SIVERTEX and poly arrays.
	for(i=0;i<1100;i++)
		verts[i].flags=FALSE;

	indent=0;

//	if(siobj_parse_obj((short*)buf.data,buf.pos/2)==FALSE)
	if(siobj_parse_obj((short*)&buf.data[buf.pos],0)==FALSE)
		{
		LogPrintf("Error loading SOD\n");
		return(FALSE);
		}
	return(TRUE);
}


Mesh *siobj_load(char *name)
{
int size;
int bi;	//buffer index.

	LogPrintf("loading:%s\n",name);

	Buffer buf;
	if(!buf.load_file(name))
		{
		LogPrintf("Can't load SIOBJ:%s\n",name);
		return(NULL);
		}


	verbose=FALSE;

	if(verbose)
      LogPrintf("Decoding:%s size:%x\n",name,buf.size);

	Mesh *obj;
	obj=new Mesh;

	obj->start_dlist();

	siobj_parse(buf);

	obj->end_dlist();
	return(obj);
}





















