// island.c

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "buffer.h"
#include "mesh.h"
#include "log.h"
#include "oglview.h"

extern unsigned char sipal[];

typedef struct{
	short num;
	short pos;
	short a,b,c,d;
	short color;
	short num_verts;
	short v[7];
}MAP_AREA;


typedef struct{
	short x,y,z;
	short f[11];
}ANCHORPT;


Mesh *island_load()
{
int i,j;
int num;
int cur_color;
int valid;
ANCHORPT *anchorpts;
MAP_AREA *map_areas;

	Buffer pts_buf;
	if(!pts_buf.load_file("sidata\\anchorpt.dat") )
		{
		LogPrintf("Can't load ANCHORPTS\n");
		return(NULL);
		}

	anchorpts=(ANCHORPT *)&pts_buf.data[0];
//	LogPrintf("\nDecoding Anchorpts:size:%d\n",size);

	if(0)
		{
		num=pts_buf.size/sizeof(ANCHORPT);
		for(i=0;i<num;i++)
			{
//			LogPrintf("I:%5d x:%5d y:%5d z:%5d ",i,anchorpts[i].x,anchorpts[i].y,anchorpts[i].z);
//			LogPrintf("A:%6d B:%6d C:%6d D:%6d E:%6d F:%6d G:%6d H:%6d\n",
//				anchorpts[i].f[0],anchorpts[i].f[1],anchorpts[i].f[2],
//				anchorpts[i].f[3],anchorpts[i].f[4],anchorpts[i].f[5],
//				anchorpts[i].f[6]);
			}
		}


	Buffer map_buf;

	if(!map_buf.load_file("sidata\\map.bin") )
		{
		LogPrintf("Can't load MAP.BIN");
		return(NULL);
		}

	map_areas=(MAP_AREA*)&map_buf.data[0];
	num=map_buf.size/sizeof(MAP_AREA);

	Mesh *obj;
	obj = new Mesh;

	obj->start_dlist();

		//create water poly
	cur_color=128+8+16;
	cur_color=255-47;
	glColor3f((float)sipal[(cur_color*3)+0]*(1.0/255.0),
				 (float)sipal[(cur_color*3)+1]*(1.0/255.0),
				 (float)sipal[(cur_color*3)+2]*(1.0/255.0) );

	glBegin(GL_POLYGON);
	   glVertex3f(-32000,-1,-32000 );
	   glVertex3f( 32000,-1,-32000 );
	   glVertex3f( 32000,-1, 32000 );
	   glVertex3f(-32000,-1, 32000 );
	glEnd();


	float scale=1.0;

	for(i=0;i<num;i++)
		{
//		LogPrintf("??:%04X ??:%05X C:%4d NV:%3d\n  ",
//			map_areas[i].num,map_areas[i].pos,map_areas[i].color,map_areas[i].num_verts);
//		LogPrintf("A:%4d B:%4d C:%4d D:%4d ",
//			map_areas[i].a,map_areas[i].b,map_areas[i].c,map_areas[i].d);

			//build the polys
		if(map_areas[i].pos&0xff)
			{
			if(cur_color!=map_areas[i].color)
				{
	         cur_color=map_areas[i].color;
				glColor3f((float)sipal[(cur_color*3)+0]*(1.0/255.0),
						 (float)sipal[(cur_color*3)+1]*(1.0/255.0),
						 (float)sipal[(cur_color*3)+2]*(1.0/255.0) );
				}


			valid=TRUE;
			for(j=0;j<map_areas[i].num_verts;j++)
				{
					//revisit: kludgy bullshit to get around TAP points.
				if(anchorpts[map_areas[i].v[j]].y>32000 ||
						anchorpts[map_areas[i].v[j]].y<-32000)
					valid=FALSE;
				}
			if(valid)
				{
				glBegin(GL_POLYGON);
				for(j=0;j<map_areas[i].num_verts;j++)
					{
				   glVertex3f(-(float)(anchorpts[map_areas[i].v[j]].x)*scale,
   	                    -(float)(anchorpts[map_areas[i].v[j]].y)*scale,
					           (float)(anchorpts[map_areas[i].v[j]].z)*scale);
					}
				glEnd();
				}

			}
		}

	obj->end_dlist();
	return(obj);

}


