#include <stdio.h>
#include <stdarg.h>
#include "sod.h"	//SOD definitions.
#include "pal.h"	//stunt island palette.

#define TRUE (1)
#define FALSE (0)

//todo
// only output used vertexes
// support subobjects
// figure out unknown opcodes
// add option to auto remove detail polys
// support lines block circle and sphere in that order.
// calculate texture verts
// create a color to texture look up table


typedef struct{
	int x,y,z;
	int other;
	int flags;
	}VERTEX;

#define MAX_POLY_VERTS	(20) //NOTE: hardcoded max of 20.
									  //I have never seen this exceeded.
typedef struct{
	int v[MAX_POLY_VERTS];	
	int num;		 //number of verts.
	int flags;
	int color;
	}POLY;

POLY polys[10000];
int poly_index=0;

VERTEX verts[10000];
VERTEX cur_vert;
int vert_index=0;
int debug=TRUE;


	//Writes out a generic wave .MTL file.
	//Materials are built from the SI palette.
	//Writes out 256 diffrent materials. One for
	//each color in the SI palette.

void write_mtl()
{
FILE *f;
int i;
float con;

	f=fopen("si.mtl","wt");

	fprintf(f,"newmtl default\n");
	fprintf(f,"\tNs 32\n");
	fprintf(f,"\td 1\n");
	fprintf(f,"\tillum 2\n");
	fprintf(f,"\tKd 0.4 0.4 0.4\n");
	fprintf(f,"\tKs 0.7 0.7 0.7\n");
	fprintf(f,"\tKa 0.3 0.3 0.3\n");

	for(i=0;i<3*256;i++)
		if(sipal[i]==0)
			sipal[i]=1;

	con=1.0/256.0;
	for(i=0;i<256;i++)
		{
		fprintf(f,"\nnewmtl color__%d\n",i);
		fprintf(f,"\tNs 6\n");
		fprintf(f,"\td 1\n");
		fprintf(f,"\tillum 2\n");

		fprintf(f,"\tKd %f %f %f\n",
				con*(float)sipal[(i*3)+0],
				con*(float)sipal[(i*3)+1],
				con*(float)sipal[(i*3)+2]);
		fprintf(f,"\tKs 0.02 0.02 0.02\n");	  //default specular
		fprintf(f,"\tKa 0.03 0.03 0.03\n");   //default ambient
		}
	fclose(f);
}

	//write a 3dstudio .ASC file.
	//Note: this fuction was abandoned. use write_wave instead.
void write_asc(char *n)
{
FILE *f;
int i;
char fn[80];

	sprintf(fn,"%s.obj",n);
	f=fopen(fn,"wt");
	fprintf(f,"Named object: \"Object01\"\n");
	fprintf(f,"Tri-mesh, Vertices: %d     Faces: %d\n",512+1,poly_index);
	fprintf(f,"Vertex list:\n");
	for(i=0;i<513;i++)
		fprintf(f,"Vertex %d:  X: %d.0     Y: %d.0     Z: %d.0\n",
				i,verts[i].x,verts[i].y,verts[i].z);

	fprintf(f,"Face list:\n");

	for(i=0;i<poly_index;i++)
		fprintf(f,"Face %d:    A:%d B:%d C:%d AB:1 BC:1 CA:1 \n"
			,i,polys[i].v[0],polys[i].v[1],polys[i].v[2]);

	fclose(f);
}

	//Write out the object as an wave .OBJ file.
	//.OBJ files are nice straight forward ASCII files that are easy
	//to convert to just about any other format.

void write_wave(char *n,float scale_factor)
{
FILE *f;
int i,v;
char fn[80];
int cur_mat=0xffff;

	sprintf(fn,"%s.obj",n);
	f=fopen(fn,"wt");

	printf("Writing Obj to %s. Polys:%d Scale:%f\n",n,poly_index,scale_factor);

	write_mtl();

	f=fopen(fn,"wt");

	fprintf(f,"mtllib si.mtl\n");
	for(i=0;i<1024;i++)
		fprintf(f,"v %f %f %f\n",
			(float)verts[i].x*scale_factor,
         (float)verts[i].y*scale_factor,
			(float)verts[i].z*scale_factor);


	fprintf(f,"# %d vertices\n",1024);

	fprintf(f,"# 0 texture vertices\n");

	fprintf(f,"# 0 normals\n");

	fprintf(f,"usemtl default\n");

	fprintf(f,"o %s\n",n);

	fprintf(f,"g %s\n",n);
	fprintf(f,"s off\n");
	for(i=0;i<poly_index;i++)
		{
		if(cur_mat!=polys[i].color)
			{
         cur_mat=polys[i].color;
      	fprintf(f,"usemtl color__%d\n",cur_mat);
			}
		fprintf(f,"f ");
		for(v=polys[i].num-1;v>-1;v--)
			fprintf(f,"%d ",polys[i].v[v]+1);
		fprintf(f,"\n");
		}
	fprintf(f,"# %d polygons\n",poly_index);

	fclose(f);
}


void debug_printf(const char *format, ...)
{
	char texbuffer[200];
	va_list argp;

	if(!debug)
		return;

	va_start(argp, format);
	vsprintf(texbuffer,format,argp);
	printf("%s",texbuffer);
}


void fatal_error(char *s)
{
	puts(s);
	exit(-1);
}
short next_word(FILE *f)
{
short w;
	fread(&w,sizeof(short),1,f);
	return(w);
}

	//C versions of the model opcodes
set_vert(VERTEX *v,int x,int y,int z)
{
	v->x=x;v->y=y;v->z=z;
	v->flags=TRUE;
}

get_vert(VERTEX *d,VERTEX *s)
{
	d->x=s->x;
	d->y=s->y;
	d->z=s->z;
	d->flags=TRUE;

}
mid_vert(VERTEX *d,VERTEX *s)
{
VERTEX t;

//	if(!s->flags || !d->flags)
//		printf("Error:reading Un-inited vert");

	t.x=(s->x-d->x)/2;
	t.y=(s->y-d->y)/2;
	t.z=(s->z-d->z)/2;
	
	d->x+=t.x;
	d->y+=t.y;
	d->z+=t.z;

}

add_vert(VERTEX *d,VERTEX *s)
{
VERTEX t;

	t.x=(s->x+d->x);
	t.y=(s->y+d->y);
	t.z=(s->z+d->z);
	
	d->x=t.x;
	d->y=t.y;
	d->z=t.z;

}


sub_vert(VERTEX *d,VERTEX *s)
{
VERTEX t;

	t.x=(d->x-s->x);
	t.y=(d->y-s->y);
	t.z=(d->z-s->z);
	
	d->x=t.x;
	d->y=t.y;
	d->z=t.z;

}
neg_vert(VERTEX *v)
{
	v->x=-v->x;
	v->y=-v->y;
	v->z=-v->z;
}
shr_vert(VERTEX *v,int s)
{
	v->x=v->x>>s;
	v->y=v->y>>s;
	v->z=v->z>>s;
}
shl_vert(VERTEX *v,int s)
{
	v->x=v->x<<s;
	v->y=v->y<<s;
	v->z=v->z<<s;
}

dump_vert(VERTEX *v)
{
	if(!v->flags)
		printf("Dumping Un-inited vert");
	debug_printf("--->%d,%d,%d",v->x,v->y,v->z);
}

main(int ac,char *av[])
{
FILE *f;
short w,w1,w2,w3,w4,w5;
int i;
int num;
int size;
char name[80];
int cur_color=0;
int scale=14;
float scale_factor;

	sprintf(name,"%s",av[1]);
	for(i=0;i<strlen(name);i++)
		if(name[i]=='.')
			name[i]=0;

	f=fopen(av[1],"rb");
	if(f==NULL)
		fatal_error("file not found");

	fseek(f,0,SEEK_END);
	size=ftell(f);
	fseek(f,0,SEEK_SET);
	printf("\nDecoding:%s size:%d\n",av[1],size);

	w=next_word(f);
	if(w!=OP_ID)
		fatal_error("Id not found");

	w=next_word(f);
	debug_printf("2nd word %x:%d\n",w,w);	  //dunno what this word is

	w=next_word(f);
	debug_printf("3rd word %x:%d\n",w,w);    //dunno what this word is

	scale=w=next_word(f);
	debug_printf("Scale %x:%d\n",w,w);
	if(scale>17)
		fatal_error("Scale out of range!");
	scale_factor=sod_scale_table[scale];

	w=next_word(f);
	debug_printf("5th word %x:%d\n",w,w);    //dunno what this word is

		//clear out vertex and poly arrays.
	for(i=0;i<1100;i++)
		verts[i].flags=FALSE;
	for(i=0;i<1100;i++)
		polys[i].flags=FALSE;

	while(ftell(f)<size)	//loop 'till end of file
		{
		w=next_word(f);
		switch(w)
			{
			case OP_IPG:	 //initilize vertex cache
				w=next_word(f);
				switch(w)
					{
					case 0:		//make a cube starting at vertex in w4
		       		w1=next_word(f);
		        		w2=next_word(f);
		        		w3=next_word(f);
		        		w4=next_word(f)/0xc;
						vert_index=w4;
						debug_printf("IPG CUBE <%d,%d,%d>,%d\n",w,w1,w2,w3,w4);
						set_vert(&verts[vert_index+7],0+w1,0+w2,0-w3);
						set_vert(&verts[vert_index+6],0+w1,0+w2,0+w3);
						set_vert(&verts[vert_index+5],0-w1,0+w2,0+w3);
						set_vert(&verts[vert_index+4],0-w1,0+w2,0-w3);
						set_vert(&verts[vert_index+3],0+w1,0-w2,0-w3);
						set_vert(&verts[vert_index+2],0+w1,0-w2,0+w3);
						set_vert(&verts[vert_index+1],0-w1,0-w2,0+w3);
						set_vert(&verts[vert_index+0],0-w1,0-w2,0-w3);
					  	break;
					case 2:  //make a square starting at vertex in w4
		       		w1=next_word(f);
		        		w2=next_word(f);
		        		w4=next_word(f)/0xc;
						vert_index=w4;
						debug_printf("IPG GROUNDPTS <%d,%d>,%d\n",w,w1,w2,w4);
						set_vert(&verts[vert_index+3],0+w1,0,0-w2);
						set_vert(&verts[vert_index+2],0+w1,0,0+w2);
						set_vert(&verts[vert_index+1],0-w1,0,0+w2);
						set_vert(&verts[vert_index+0],0-w1,0,0-w2);
					  	break;
					case 8:  //make a square starting at vertex in w4
		       		w1=next_word(f);
		        		w2=next_word(f);
		        		w4=next_word(f)/0xc;
						vert_index=w4;
						debug_printf("IPG FLAT <%d,%d>,%d\n",w,w1,w2,w4);
						set_vert(&verts[vert_index+3],0+w1,0-w2,0);
						set_vert(&verts[vert_index+2],0+w1,0+w2,0);
						set_vert(&verts[vert_index+1],0-w1,0+w2,0);
						set_vert(&verts[vert_index+0],0-w1,0-w2,0);
					  	break;
					default:
						printf("IPG type:%d\n",w);
						fatal_error("IPG type wrong");
					  	break;
					}
				break;
			case OP_VECTORS:
				w=next_word(f);			//number of verts
				w1=next_word(f)/0xc;		//start register.
				debug_printf("VECTORS #%d Start:%d\n",w,w1);
				for(i=0;i<w;i++)
					{
					w2=next_word(f);
					w3=next_word(f);
					w4=next_word(f);

					//HMM not sure why the -w3 clipper mast bug fix
               set_vert(&verts[w1],w2,-w3,w4);

					debug_printf("VEC %d ",w1);
					dump_vert(&verts[w1]);
					debug_printf("\n");
					w1++;
					}
				break;
			case OP_GETV:
				w=next_word(f)/0xc;
				get_vert(&cur_vert,&verts[w]);
				debug_printf("GETV %d ",w);
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_INTV:
				w=next_word(f)/0xc;
				mid_vert(&cur_vert,&verts[w]);
				debug_printf("INTV %d ",w);
				dump_vert(&verts[w]);
				debug_printf("\n");
				break;
			case OP_PUTV:
				w=next_word(f)/0xc;
				get_vert(&verts[w],&cur_vert);
				debug_printf("PUTV %d ",w);
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_ADDV:
				w=next_word(f)/0xc;
				add_vert(&cur_vert,&verts[w]);
				debug_printf("ADDV %d ",w);
				dump_vert(&verts[w]);
				debug_printf("\n");
				break;
			case OP_SUBV:
				w=next_word(f)/0xc;
				sub_vert(&cur_vert,&verts[w]);
				debug_printf("SUBV %d ",w);
				dump_vert(&verts[w]);
				debug_printf("\n");
				break;
			case OP_SHRV:
				w=next_word(f);
				shr_vert(&cur_vert,w);
				debug_printf("SHRV %d ",w);
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_SHLV:
				w=next_word(f);
				shl_vert(&cur_vert,w);
				debug_printf("SHLV %d ",w);
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_NEGV:
				neg_vert(&cur_vert);
				debug_printf("NEGV ");
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_CLRV:
				set_vert(&cur_vert,0,0,0);
				debug_printf("CLRV");
				dump_vert(&cur_vert);
				debug_printf("\n");
				break;
			case OP_PERS:
				w=next_word(f);
				w1=next_word(f);
				debug_printf("PERSPEC %d,%d\n",w1,w);
				break;
			case OP_REJ:
				w=next_word(f);
				debug_printf("REJ3D %d\n",w);
				w=next_word(f);
				debug_printf("Num Faces:%d\n",w);
				for(i=0;i<w;i++)
					{
           		w1=next_word(f);
           		w2=next_word(f);
           		w3=next_word(f);
					debug_printf("face <%d,%d,%d>\n",w1/0xc,w2/0xc,w3/0xc);
					}
				break;
			case OP_VEC:
        		w1=next_word(f);
           	w2=next_word(f);
           	w3=next_word(f);
				debug_printf("vec <%d,%d,%d>\n",(int)w1,(int)w2,(int)w3);
				break;
			case OP_SPHERE:  //primitive not supported
        		w1=next_word(f);
        		w2=next_word(f);
				debug_printf("SPHERE %d,%d\n",(int)w1,(int)w2);
				break;
			case OP_BLOCK:    //primitive not supported
        		w1=next_word(f);
        		w2=next_word(f);
				debug_printf("BLOCK %d,%d\n",(int)w1,(int)w2);
				break;
			case OP_PERFDISK: //primitive not supported
        		w1=next_word(f);
        		w2=next_word(f);
        		w3=next_word(f);
				debug_printf("PERFDISK %d,%d,%d\n",(int)w1,(int)w2,(int)w3);
				break;
			case OP_IFVIS:
				w=next_word(f);
				debug_printf("IFVIS %d\n",w);
				w=next_word(f); //section size
//				printf("Size %x:%d\n",w);
				break;
			case OP_COLOR:
				w=next_word(f);
				cur_color=w;
				debug_printf("FILLCOLOR %d\n",w);
				break;
			case OP_LINECOL:
				w=next_word(f);
				debug_printf("LINECOL %d\n",w);
				break;
			case OP_MOVETO:
				w=next_word(f);
				debug_printf("MOVETO %d\n",w/0xc);
				break;
			case OP_ZLINETO:
				w=next_word(f);
				debug_printf("ZLINETO %d\n",w/0xc);
				break;
			case OP_DOT:
				w=next_word(f);
				debug_printf("DOT %d\n",w/0xc);
				break;
			case OP_POLYZ:
				debug_printf("POLYZ ");
				w=next_word(f);
				debug_printf("%d,",w);
				debug_printf("<");
				polys[poly_index].num=0;
				polys[poly_index].flags=TRUE;
				polys[poly_index].color=cur_color;
				for(i=0;i<w;i++)
					{
           		w1=next_word(f)/0xc;
					debug_printf(" %d",w1);
               polys[poly_index].v[polys[poly_index].num++]=w1;
//					dump_vert(&verts[w1]);
					}
				debug_printf(">\n");
				poly_index++;
				break;
			case OP_GVPOLY16:
				debug_printf("-->");
			case OP_GVPOLY:
				debug_printf("GVPOLY ");
				num=next_word(f);
				debug_printf("%d,%d",num,w);
				debug_printf("<");
				polys[poly_index].num=0;
				polys[poly_index].flags=TRUE;
				polys[poly_index].color=cur_color;
				for(i=0;i<num;i++)
					{
           		w1=next_word(f)/0xc;
               polys[poly_index].v[polys[poly_index].num++]=w1;
					debug_printf("%d,",w1);
					}
				poly_index++;
				w=next_word(f)/0xc;
				debug_printf(">,%d,<",w);
				for(i=0;i<num;i++)
					{
           		w1=next_word(f);
					printf("%d,",w1/2);
					}
				debug_printf(">\n");
				break;
			case OP_END:
				debug_printf("ENDSEC\n");
				break;
			case OP_SHADES:
				w=next_word(f);
				debug_printf("SHADES %d\n",w);
				w=next_word(f);
				for(i=0;i<w;i++)
					{
	        		w1=next_word(f);
   	        	w2=next_word(f);
   	        	w3=next_word(f);
					debug_printf("vec <%d,%d,%d>\n",(int)w1,(int)w2,(int)w3);
					}
				break;
			case 0x18c:
				w1=next_word(f);
				w2=next_word(f);
				w3=next_word(f);
				debug_printf("SORT %d %d %d\n",w1,w2,w3);
				break;
			default:
				w1=next_word(f);
				printf("%05X:Unknown op:%x:%d next %x:%d\n",ftell(f),w,w,w1,w1);
				printf("Writing object anyway. Obj may be corrupt.\n");
				write_wave(name,scale_factor);
				exit(-1);
				break;

			}

		}
	fclose(f);

//	write_asc("out.asc");
	write_wave(name,scale_factor);

}















